<?php
namespace SupplyTek\Controllers;

use Zeedhi\Framework\DTO\Request;
use Zeedhi\Framework\DTO\Response;
use SupplyTek\Utilities\CrudSupplyTek;
use SupplyTek\Utilities\SessionHandler;
use SupplyTek\Services\OrdersService;

/**
 * Returns information about orders
 *
 * @author Grazielle Ferreira Antunes
 */
class OrdersController 
{
    protected $crud;
    protected $sessionHandler;
    protected $ordersService;
    protected $fileServerURL;
    protected $zhFileUploader;
    protected $supBaseUri;
    

    /**
     * Class constructor
     * 
     * Initialize an object's properties upon creation of the object.
     * 
     * @param CrudSupplyTek $crudSupplyTek Object that allows to perform queries from datasource
     * @param SessionHandler $sessionHandler Object that allows to manipulate session
     * @param OrdersService $ordersService Service responsible for handling the database
     * @param $fileServerUrl Handles the URL of images and files
     * @param $zhFileUploader Run FileServer methods
     * @param $supBaseUri BaseUri from SUP
     * 
     * @return void
     * 
     * @access public
     * @author Grazielle Ferreira Antunes
     */
    public function __construct($crudSupplyTek, $sessionHandler, $ordersService, $fileServerURL, $zhFileUploader, $supBaseUri) 
    {
        $this->crud = $crudSupplyTek;
        $this->sessionHandler = $sessionHandler;
        $this->ordersService = $ordersService;
        $this->fileServerURL = $fileServerURL;
        $this->zhFileUploader = $zhFileUploader;
        $this->supBaseUri = $supBaseUri;
    }

    /**
     * It prepares the search data for the service and 
     * returns the information of the last 6 requests to the front end.
     *
     * @param Request $request Object that receives data from the caller.
     * @param Response $response Object that sends back data to the caller.
     * 
     * @return array
     * @throws Exception
     *
     * @access public
     * @author Grazielle Ferreira Antunes
     */
    public function getLastOrders(Request $request, Response $response)
    {
        try {
            $unitId = $this->sessionHandler->getSessionData('CDFILIAL');
            $orgId = $this->sessionHandler->getSessionData('NRORG');
            
            $params = array(
                "UNIT_ID" => $unitId,
                "ORG_ID" => $orgId
            );
            
            $filterCriteria = $request->getFilterCriteria();
            $filter = $this->crud->processFilter($filterCriteria);
            
            if (!isset($filter["SUPPLIER_ID"])) {
                $params["SUPPLIER_ID"] = "T";                
            }
            
            $lastOrders = $this->crud->find($request, $params);

            $response->setMetadata($lastOrders);
        } catch (\Exception $e) {
            throw $e;
        }
    }
    
    /**
     * Return the orders filtered by status type
     * 
     * Receives a status type from request and use it to filter
     * the orders that will be shown on the grid.
     * 
     * @param Request $request Object that receives data from the caller
     * @param Response $response Object that sends back data to the caller
     * 
     * @return array
     * @throws Exception
     * 
     * @access public
     * @author Rômulo Alves Lousada
     */
     public function getOrdersByStatus(Request $request, Response $response)
     {
         try {
            $perfilId =  $this->sessionHandler->getSessionData('IDPERFILACES');
            $unitId = $this->sessionHandler->getSessionData('CDFILIAL');
            $operatorId = $this->sessionHandler->getSessionData('CDOPERADOR');
            $filterUri = $request->getParameters();
            $filterCriteria = $request->getFilterCriteria();

            $filter = $this->crud->processFilter($filterCriteria, $filterUri);
    
            $params = array(
                "OPERATOR_ID" => $operatorId
            );
            
            if (!isset($filter["BRANCH_ID"])) {
                $params["BRANCH_ID"] = $unitId;
            }

            if (!isset($filter["STATUS"])) {
                $params["STATUS"] = 'T';
            }

            if (!isset($filter["ID"])) {
                $params["ID"] = 'T';
            }
            
            if($perfilId == 'GO') {
                $filterCriteria->setDataSourceName("orders-GO"); 
            }
            
            $ordersByStatus = $this->crud->find($request, $params);
            
            $response->setMetadata($ordersByStatus);
        } catch (\Exception $e) {
            throw $e;
        }
         
     }
     
    /**
     * Return the number of pending orders and the newest date
     * 
     * Searches the database counting the number of pending orders for the
     * logged unit and the newest date of those orders, to show on the starter
     * page.
     * 
     * @param Request $request Object that receives data from the caller
     * @param Response $response Object that sends back data to the caller
     * 
     * @return array
     * @throws Exception
     * 
     * @access public
     * @author Rômulo Alves Lousada
     */
     public function getPendingOrders(Request $request, Response $response)
     {
         try {
            $operatorId = $this->sessionHandler->getSessionData("CDOPERADOR");
            
            $filter = $this->crud->processFilter($request->getFilterCriteria());
            
            $params = array(
                "OPERATOR_ID" => $operatorId
            );
            
            if (!isset($filter["BRANCH_ID"])) {
                $params["BRANCH_ID"] = 'T';                
            }
            
            $pendingOrders = $this->crud->find($request, $params);
            
            $response->setMetadata($pendingOrders);
        } catch (\Exception $e) {
            throw $e;
        }
         
     }
     
    /**
     * Return details from itens inside an order
     * 
     * Retrieves detailed data from items of an order to be shown.
     *
     * @param Request $request Object that receives data from the caller.
     * @param Response $response  Object that sends back data to the caller.
     *
     * @return array
     * @throws Exception
     * 
     * @access public
     * @author Rômulo Alves Lousada
     */
    public function getItemsFromOrder(Request $request, Response $response)
    {
        try {
            $itemsFromOrder = $this->crud->find($request);
            $newDataSet = $this->setImgFullPath($itemsFromOrder);
            $response->setMetadata($newDataSet);
        } catch (\Exception $e) {
            throw $e;
        }
    }
    
    /**
     * Returns full image address
     * 
     * Get the product image link on the file server and link it with the image 
     * link brought from the bank so that the image is displayed on the system.
     *
     * @param string $products Array with order products
     * 
     * @return array
     * 
     * @acess private
     * @author Bruna Chaves Carneiro
     */
	private function setImgFullPath($products) 
	{
        if(!empty($products['data'])) {
    		foreach ($products['data'] as &$product) {
    			if ($product['IMAGE']) {
    				$product['IMAGE'] = $this->fileServerURL . $product['IMAGE'];
    			} 
    		}
        }
		return $products;
	}
    
    /**
     * Main function from OrdersSummary route that process request and provide response
     * 
     * Receives the request sent by the frontend, execute validations and 
     * calculate the percentage of amount orders in period.
     * 
     * @param Request $request Object that receives data from the caller
     * @param Response $reponse Object that provides data to whoever requested
     *
     * @return Object
     * 
     * @access public
     * @author Danilo D. Vaz
     */
    public function getOrdersSummary (Request $request, Response $response)
    {
        try {
            $filterCriteria = $request->getFilterCriteria();
            $params['OPERATOR_ID'] = $this->sessionHandler->getSessionData('CDOPERADOR');
            
            $filter = $this->crud->processFilter($filterCriteria);
            $params['BRANCH'] = $this->getBranchOrdersSummary($filter);
            list($params['FIRST_DAY'], $params['LAST_DAY']) = $this->getPeriodOrdersSummary($filter);
            
            $ordersSummary = $this->crud->executeQuery('orders-summary', $params);
            
            $totalOrders = $this->getTotalOrdersSummary($ordersSummary);
            
            $dataOrdersSummary = $this->buildDataOrdersSummary($ordersSummary, $totalOrders);

            $data = $this->crud->buildData(array($dataOrdersSummary), $filterCriteria);
            
            $response->setMetadata($data);  
        } catch (\Exception $e) {
            throw new \Exception ($e->getMessage(), 500);
        }
    }
    
    /**
     * Function from OrdersSummary route that process unit 
     * 
     * Receives the filter sent by frontend and checks the unit. If not informed
     * any unit, returns the session unit.
     * 
     * @param array $filter Array with filter sent by frontend
     *
     * @return string
     * 
     * @access private
     * @author Danilo D. Vaz
     */
    private function getBranchOrdersSummary ($filter)
    {
        if (isset($filter['BRANCH'])) {
            return $filter['BRANCH'];
        }
       
        return $this->sessionHandler->getSessionData('CDFILIAL');
    }
    
    /**
     * Function from OrdersSummary route that process period 
     * 
     * Receives the filter sent by frontend and checks the period. If not informed
     * any period, returns the current month. Execute other validations in period.
     * Returns the first day and the last day of the month.
     * 
     * @param array $filter Array with filter sent by frontend
     *
     * @return array
     * @throws Exception [if can't create date object]
     * 
     * @access private
     * @author Danilo D. Vaz
     */
    private function getPeriodOrdersSummary ($filter)
    {
        if (isset($filter['PERIOD'])) {
            $date = $filter['PERIOD'];
        } else {
            $date = $period = date('m/y');
        }
        
        $date = \DateTime::createFromFormat('m/y', $date);
        
        if (!$date) {
            throw new \Exception("Wrong date format.");
        }

        $date->modify('first day of this month');
        $firstDay = $date->format('d/m/Y');
        
        $date->modify('last day of this month');
        $lastDay = $date->format('d/m/Y');
        
        return [$firstDay, $lastDay];
    }
    
    /**
     * Function from OrdersSummary route that sum the orders amount 
     * 
     * Receives the query result and calculate tem total orders amount.
     * 
     * @param array $ordersSummary Array with query result
     *
     * @return float
     * 
     * @access private
     * @author Danilo D. Vaz
     */
    private function getTotalOrdersSummary ($ordersSummary)
    {
        $totalOrders = 0;
        
        foreach ($ordersSummary as $orders) {
            $totalOrders += $orders['QTTOTAF'];
        }
        
        return $totalOrders;
    }
    
    /**
     * Function from OrdersSummary route that build data
     * 
     * Receives the query result and total orders amount. Build a data array
     * in the default expected by the frontend.
     * 
     * @param array $ordersSummary Array with query result
     * @param float $totalOrders Float with total orders amount
     *
     * @return array
     * 
     * @access private
     * @author Danilo D. Vaz
     */
    private function buildDataOrdersSummary ($ordersSummary, $totalOrders)
    {
        $dataFormat = self::dataFormatOrdersSummary();
        
        if (count($ordersSummary)) {
            foreach ($ordersSummary as $order) {
                switch ($order['IDSITAFALCADA']) {
                    case 'APROVADA':
                        $dataFormat['APPROVED']['VALUE'] = round($order['VRTOTAF'], 2);
                        $dataFormat['APPROVED']['PERCENTAGE'] = $this->calcPercentOrdersSummary($order['QTTOTAF'], $totalOrders);
                    break;
                    
                    case 'PENDENTE':
                        $dataFormat['PENDING']['VALUE'] = round($order['VRTOTAF'], 2);
                        $dataFormat['PENDING']['PERCENTAGE'] = $this->calcPercentOrdersSummary($order['QTTOTAF'], $totalOrders);
                    break;
                    
                    case 'RECUSADA':
                        $dataFormat['REJECTED']['VALUE'] = round($order['VRTOTAF'], 2);
                        $dataFormat['REJECTED']['PERCENTAGE'] = $this->calcPercentOrdersSummary($order['QTTOTAF'], $totalOrders);
                    break;
                }
            }
        }
        
        return $dataFormat;
    }
    
    /**
     * Function from OrdersSummary route that provides default data
     * 
     * Returns the default data.
     * 
     * @return array
     * 
     * @access private
     * @author Danilo D. Vaz
     */
    private static function dataFormatOrdersSummary ()
    {
        $ordersSituation = array(
            "APPROVED" => array(
                "VALUE" => 0,
                "PERCENTAGE" => 0
            ),
            "PENDING" => array(
                "VALUE" => 0,
                "PERCENTAGE" => 0
            ),
            "REJECTED" => array(
                "VALUE" => 0,
                "PERCENTAGE" => 0
            )
        );
        
        return $ordersSituation;
    }
    
    /**
     * Function from OrdersSummary route that calculate percentage
     * 
     * Receives the orders amount and the total orders amount, and calculate
     * the percentage. Returns the value with two decimal places.
     * 
     * @param float $amount Float with orders amount
     * @param float $totalOrders Float with total orders amount
     *
     * @return float
     * 
     * @access private
     * @author Danilo D. Vaz
     */
    private function calcPercentOrdersSummary ($amount, $totalOrders)
    {
        $percentage = 0;
        
        if ($amount && $totalOrders) {
            $percentage = (100 * $amount) / $totalOrders;
        }
        
        return round($percentage, 2);
    }

    /**
     * Retrieves available quantity, quantity effected and average approval period
     * 
     * Obtain the available budget amount, the amount spent on orders placed, and the average approval period
     * per unit, according to the period.
     * 
     * @param Request $request Object that receives data from the caller
     * @param Response $response Object that sends back data to the caller
     *
     * @return array
     * @throws Exception
     * 
     * @access public
     * @author Bruna Chaves Carneiro
     */
    public function getOrdersMonitoring(Request $request, Response $response)
    {
        try {
            $filterCriteria = $request->getFilterCriteria();
            $branch = $this->sessionHandler->getSessionData('CDFILIAL');
            $period = date('d/m/Y');
            $params = array(
                "BRANCH"   => $branch,
                "PERIOD"   => $period
            );
            
            $ordersMonitoring = $this->crud->executeQuery('orders-monitoring', $params);
            $data = $this->crud->buildData($ordersMonitoring, $filterCriteria);
            
            $response->setMetadata($data);
            
        } catch (\Exception $e) {
            throw new \Exception ($e->getMessage(), 500);
        }
    }
    
    /**
     * Returns order approvers;
     * 
     * Receives the order number as a parameter and returns the approvers of 
     * that order according to the branch registered in the system.
     *
     * @param Request $request Object that receives data from the caller
     * @param Response $response Object that sends back data to the caller
     * 
     * @return array
     * @throws Exception
     * 
     * @access public
     * @author Bruna Chaves Carneiro
     */
    public function getOrderApprovals(Request $request, Response $response)
    {
        try {
            $cdFilial = $this->sessionHandler->getSessionData('CDFILIAL');
            
            $params = array(
                "P_CDFILDEST"  => $cdFilial
            );
            
            $orderApprovals = $this->crud->find($request, $params);
            
            $response->setMetadata($orderApprovals);
        } catch (\Exception $e) {
            throw $e;
        }
    }
    
    /**
     * Returns order comments
     * 
     * Receives the order number as a parameter and returns the comments for that
     * order according to the branch registered in the system.
     *
     * @param Request $request Object that receives data from the caller
     * @param Response $response Object that sends back data to the caller
     *
     * @return array
     * @throws Exception
     * 
     * @access public
     * @author Bruna Chaves Carneiro
     */
    public function getOrderComments(Request $request,Response $response)
    {
        try {
            $cdFilial = $this->sessionHandler->getSessionData('CDFILIAL');

            $params = array(
                "P_CDFILDEST"  => $cdFilial
            );
            
            $orderComments = $this->crud->find($request, $params);
            $this->setArchiveFullPath($orderComments);
            
            $response->setMetadata($orderComments);
        } catch (\Exception $e) {
            throw $e;
        }
    }
    
    /**
     * Returns full archive address
     * 
     * Get the download file link
     *
     * @param array $orderComments Array with order comments
     * 
     * @return void
     * 
     * @acess private
     * @author Grazielle Ferreira Antunes
     */
    private function setArchiveFullPath(&$orderComments) 
	{
        if(!empty($orderComments['data'])) {
    		foreach ($orderComments['data'] as &$orderComments) {
    			if ($orderComments['ATTACHMENT_URL']) {
    				$orderComments['ATTACHMENT_URL'] = $this->fileServerURL . $orderComments['ATTACHMENT_URL'];
    			} 
    		}
        }
	}
    
    /**
     * Function that refuse order
     *
     * @param Request $request Object that receives data from the caller.
     * @param Response $response Object that sends back data to the caller.
     * 
     * @return void
     * @throws Exception
     *
     * @access public
     * @author Grazielle Ferreira Antunes
     */
    public function refuseOrder(Request $request, Response $response)
    {
        try {
            $row = $request->getDataSet()->getRows()[0];
            $status = 'R';
            $this->prepareParams($row, $status);
        } catch (\Exception $e) {
            throw $e;
        }
    }
    
    /**
     * Function that approve order
     *
     * @param Request $request Object that receives data from the caller.
     * @param Response $response Object that sends back data to the caller.
     * 
     * @return void
     * @throws Exception
     *
     * @access public
     * @author Grazielle Ferreira Antunes
     */
    public function approveOrder(Request $request, Response $response)
    {
        try {
            $row = $request->getDataSet()->getRows()[0];
            $status = 'A';
            $this->prepareParams($row, $status);
        } catch (\Exception $e) {
            throw $e;
        }
    }
    
    /**
     * Prepares the parameters to start the approval or refusal process 
     * of the order and calls the method that executes this process.
     * 
     * @param array $row Order information.
     * 
     * @return array
     *
     * @access private
     * 
     * @author Grazielle Ferreira Antunes
     */
    private function prepareParams($row, $status)
    {
        $operatorId = $this->sessionHandler->getSessionData('CDOPERADOR');
        $branchId = $this->sessionHandler->getSessionData('CDFILIAL');
        $organizationId = $this->sessionHandler->getSessionData('NRORG');
    
        $params = array(
            "CDFILDEST"     => $row['order']['BRANCH'],
            "CDFILLANC"     => $row['order']['CDFILLANC'],
            "NRAUTOFORN"    => $row['order']['ID'],
            "CDOPERADOR"    => $operatorId,
            "STATUS"        => $status,
            "DSJUSTIFIC"    => $row['justification'],
            "NRAPROVALCADA" => $row['order']['NRAPROVALCADA'],
            "VRAPROVCMP"    => $row['order']['VRAPROVCMP'],
            "CDFILIAL"      => $branchId,
            "NRORG"         => $organizationId,
            "SUPPLYTEK_URL" => "http://" . $_SERVER['HTTP_HOST']
        );

        $this->executeWorkflow($params);
    }
    
    
    /**
     * Receives the parameters to start the process of approval or refusal 
     * of the order and calls the method on a public route of the supplies 
     * module at Tecfood.
     * 
     * @param array $params Parameters required for approval or refusal of the order
     *
     * @access private
     * 
     * @author Grazielle Ferreira Antunes
     */
    private function executeWorkflow($params)
    {
        $dsdns = $this->crud->executeQuery('appConfig');

        if (empty($dsdns[0]['DSDNS'])) {
            throw new \Exception("Tecfood address not found");
        } else {
            $dsdns = $dsdns[0]['DSDNS'];
        }

        $url = $dsdns . $this->supBaseUri . '/supplyTekApproval';
        $data = array(
            'requestType' => 'Row',
            'row' => $params
        );

        $header[] = 'Content-Type: application/json';
        $header[] = 'Content-Length: ' . strlen(json_encode($data));
        
        $ch = curl_init();
        
        curl_setopt($ch, CURLOPT_HTTPHEADER, $header);
        curl_setopt($ch, CURLOPT_URL, $url);
        curl_setopt($ch, CURLOPT_POST, count($data));
        curl_setopt($ch, CURLOPT_POSTFIELDS, json_encode($data));
        curl_setopt($ch, CURLOPT_CUSTOMREQUEST, "POST");
        curl_setopt($ch, CURLOPT_RETURNTRANSFER, 1);
        curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, 0);
    
        $result = curl_exec($ch);
        
        curl_close($ch);

        if (!$result || strpos($result, '404') !== false) {
            throw new \Exception("Tecfood not found at the address: '$dsdns'");
        }

        if (strlen($result) > 2) {
            throw new \Exception($result);
        }
    }
    
    /**
     * Prepares the parameters and calls the service method 
     * to insert the comment in the order.
     * 
     * @param Request $request Object that receives data from the caller.
     * @param Response $response Object that sends back data to the caller.
     * 
     * @access public
     * 
     * @author Grazielle Ferreira Antunes
     */
    public function sendOrderComment (Request $request, Response $response)
    {
        try {
            $this->ordersService->beginTransaction();
            
            $row = $request->getDataSet()->getRows()[0];
            
            $nrOrg = $this->sessionHandler->getSessionData('NRORG');
            
            if(isset($row['formData']['commentFiles'])){
                $fileRow = $this->postFileServer($row['formData']['commentFiles']);
            }else{
                $fileRow['UPLOADCOMMENT'][0] = null;
            }

            $this->ordersService->insertOrderComment($row['formData'], $row['order'], $fileRow['UPLOADCOMMENT'][0], $nrOrg);
            
            $this->ordersService->commit();
        } catch (\Exception $e) {
            $this->ordersService->rollback();
            throw $e;
        } 
        
    }
    
    /**
     * Upload the file to FileServe.
     * 
     * @param array $rowFile  User-uploaded file.
     * 
     * @return array
     *
     * @access private
     * 
     * @author Grazielle Ferreira Antunes
     */
    private function postFileServer ($rowFile) {
        
        $fileRow = array(
            'UPLOADCOMMENT' => array(
                array(
                    'b64File' => $rowFile['b64File'],
                    'name' => $rowFile['name'],
                    'type' => $rowFile['type'],
                    'webkitRelativePath' => '',
                    'size' => $rowFile['size']
                )
            )
        );
              
        $this->zhFileUploader->uploadFilesInRow($fileRow, 'UPLOADCOMMENT' , 'OrderComment');
        
        return $fileRow;
    }
    
    /**
     * Receives parameters, performs validations and calls 
     * the service to update the product related to the order.
     * 
     * @param Request $request Object that receives data from the caller.
     * @param Response $response  Object that sends back data to the caller.
     *
     * @access public
     * 
     * @author Grazielle Ferreira Antunes
     */
    public function updateItemsFromOrder (Request $request, Response $response){
        try {
            $this->ordersService->beginTransaction();
            
            $row = $request->getDataSet()->getRows()[0];
            
            $nrOrg = $this->sessionHandler->getSessionData('NRORG');
            if($row['QUANTITY'] <= $row['QUANTITY_ORIGINAL']){
                $this->ordersService->updateItemsFromOrder($row, $nrOrg);
            } else {
                throw new \Exception ('The product amount can not be bigger than original amount');
            }
            
            $this->ordersService->commit();
        } catch (\Exception $e) {
            $this->ordersService->rollback();
            throw $e;
        }  
    }
}
